# Amahi Home Server
# Copyright (C) 2007-2010 Amahi
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License v3
# (29 June 2007), as published in the COPYING file.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# file COPYING for more details.
# 
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the Amahi
# team at http://www.amahi.org/ under "Contact Us."

require 'command'
require 'tempfile'
require 'system-utils'
require 'amahi-api'
require 'digest/md5'
require 'platform'

class App < ActiveRecord::Base

	APP_PATH = RAILS_ENV == "development" ? "/tmp/app/%s" : "/var/hda/apps/%s"
	WEBAPP_PATH = RAILS_ENV == "development" ? "/tmp/web-apps/%s" : "/var/hda/web-apps/%s"
	INSTALLER_LOG = "/var/log/amahi-app-installer.log"
	HDA_DOWNLOAD_CACHE = "/tmp/amahi-download-cache"
	AMAHI_DOWNLOAD_CACHE_SITE = 'http://mirror.amahi.org'

	class SHA1VerificationFailed < Exception; end
	class TooManyRedirects < Exception; end

	belongs_to :webapp, :dependent => :destroy
	belongs_to :theme, :dependent => :destroy
	belongs_to :db, :dependent => :destroy
	belongs_to :server, :dependent => :destroy
	belongs_to :share, :dependent => :destroy

	has_many :app_dependencies, :dependent => :destroy
	has_many :children, :class_name => "AppDependency", :foreign_key => 'dependency_id'
	has_many :dependencies, :through => :app_dependencies

	named_scope :installed, :conditions => { :installed => true }
	named_scope :in_dashboard, :conditions => [ "installed = ? and show_in_dashboard = ?", true, true]
	named_scope :latest_first, :order => 'updated_at desc'

	before_destroy :before_destroy_hook

	validates_presence_of :name
	validates_uniqueness_of :identifier

	def initialize(identifier, app=nil)
		super()
		if app.nil?
			AmahiApi::api_key = Setting.get("api-key")
			app = AmahiApi::App.find(identifier)
		end
		# online_app_to_local_app(app)
		self.name = app.name
		self.screenshot_url = app.screenshot_url
		self.identifier = app.id
		self.description = app.description
		self.version = app.version
		self.app_url = app.app_url
		self.logo_url = app.logo_url
		self.status = app.status
		self.installed = false
	end

	def install_start
		App.transaction do
			self.installed = false
			self.install_status = 0
			self.save!
		end
		p = Process.fork do
			App.transaction { install_bg }
			return
		end
		Process.detach(p)
	end

	def uninstall_start
		p = Process.fork do
			App.transaction { uninstall_bg }
			return
		end
		Process.detach(p)
	end

	def self.install_via_daemon(identifier)
		# run the kickoff script
		cmd = File.join(RAILS_ROOT, "script/install-app --environment=#{RAILS_ENV} #{identifier} >> #{INSTALLER_LOG} 2>&1 &")
		c = Command.new cmd
		c.execute
	end


	def uninstall_via_daemon
		# run the kickoff script
		cmd = File.join(RAILS_ROOT, "script/install-app -u --environment=#{RAILS_ENV} #{self.identifier} >>  #{INSTALLER_LOG} 2>&1 &")
		c = Command.new cmd
		c.execute
	end

	def self.available
		AmahiApi::api_key = Setting.get("api-key")
		begin
			av = AmahiApi::App.find(:all).map do |online_app|
				a = App.find_by_identifier online_app.id
				if a
					nil
				else
					App.new(online_app.id, online_app)
				end
			end
			av.compact
		rescue
			[]
		end
	end

	def install_message
		# NOTE: make sure these messages match the stages below
		App.installation_message self.install_status
	end

	def self.installation_message(percent)
		case percent
			when   0: "Preparing to install ..."
			when  10: "Retrieving application information ..."
			when  20: "Installing app dependencies ..."
			when  30: "Installing package dependencies ..."
			when  40: "Downloading application and unpacking it ..."
			when  60: "Installing the application in your HDA ..."
			when  70: "Creating associated server in your HDA ..."
			when  80: "Saving application settings ..."
			when 100: "Application installed."
			else "Application message unknown at #{percent}% install."
		end
	end

	def uninstall_message
		# NOTE: make sure these messages match the stages below
		case self.install_status
			when 100: "Preparing to uninstall ..."
			when  80: "Retrieving application information ..."
			when  60: "Running uninstall scripts ..."
			when  40: "Removing application files ..."
			when  20: "Uninstalling application ..."
			when   0: "Application uninstalled."
			else "Application message unknown at #{self.install_status}% uninstall."
		end
	end

	def install_status
		App.installation_status(self.identifier)
	end

	def self.installation_status(identifier)
		status = Setting.find_by_kind_and_name(identifier, 'install_status')
		return 0 unless status
		status.reload
		status.value.to_i
	end

	def install_status=(value)
		status = Setting.find_by_kind_and_name(self.identifier, 'install_status')
		if value.nil?
			status.destroy if status
			return nil
		end
		if status 
			status.update_attribute(:value, value.to_s)
			status.reload
			return value
		end
		# create it dynamically if it does not exist
		status = Setting.create(:kind => self.identifier, :name => 'install_status', :value => value.to_s)
		status.reload
		value
	end

	def has_dependents?
		children != []
	end

	def install_bg
		initial_path = Dir.pwd
		begin
			status = Setting.create(:kind => self.identifier, :name => 'install_status', :value => "0")
			# see the install_message method for the meaning of the messages
			self.install_status = 0
			AmahiApi::api_key = Setting.get("api-key")
			self.install_status = 10
			installer = AmahiApi::AppInstaller.find identifier
			self.install_status = 20
			self.install_app_deps installer if installer.app_dependencies
			self.install_status = 30
			self.install_pkg_deps installer if installer.pkg_dependencies
			self.install_pkgs installer if installer.pkg
			app_path = APP_PATH % identifier
			mkdir app_path
			webapp_path = nil
			self.install_status = 40
			unless installer.url_name.nil?
				(name, webapp_path) = self.install_webapp installer
				self.show_in_dashboard = true
			else
				self.show_in_dashboard = false
			end
			self.create_db(:name => installer.database) if installer.database && !installer.database.blank?
			# if it has a share, create it and install it
			if installer.share
				sh = Share.find_by_comment installer.share
				if sh
					# FIXME: autohook to it. this is for legacy shares. not needed in new installs
					self.share = sh
				else
					c = installer.share.capitalize
					p = Share.full_path(installer.share)
					# FIXME - use a relative path and not harcode the share path here?
					self.create_share(:comment => c, :path => p, :rdonly => false, :visible => true, :tags => "")
					cmd = Command.new("chmod 777 #{p}")
					cmd.execute
				end
			end
			self.install_status = 60
			# if it has a server, install it and associate it
			if installer.server
				s = Server.find_by_name installer.server
				if s
					# FIXME: autohook to it. this is for legacy servers. not needed in new installs
					self.server = s
				else
					self.create_server(:name => installer.server, :comment => "#{self.name} Server")
				end
			end
			self.install_status = 70
			self.create_webapp(:name => name, :path => webapp_path, :deletable => false, :custom_options => installer.webapp_custom_options, :kind => installer.kind)
			self.theme = self.install_theme(installer) if installer.kind == 'theme'
			# run the script
			initial_user = installer.initial_user
			initial_password = installer.initial_password
			if installer.install_script
				if webapp_path
					Dir.chdir(webapp_path) do
						SystemUtils.run_script(installer.install_script, name, hda_environment(initial_user, initial_password))
					end
				else
					Dir.chdir(app_path) do
						SystemUtils.run_script(installer.install_script, name, hda_environment(initial_user, initial_password))
					end
				end
			end
			self.install_status = 80
			self.forum_url = installer.forum_url
			self.initial_user = installer.initial_user
			self.initial_password = installer.initial_password
			self.special_instructions = installer.special_instructions
			# mark it as installed
			self.installed = true
			self.save!
			self.reload
			self.install_status = 100
			Dir.chdir(initial_path)
		rescue Exception => e
			self.install_status = nil
			Dir.chdir(initial_path)
			raise e
		end
	end

	def uninstall_bg
		app_path = APP_PATH % identifier
		begin
			self.install_status = 100
			AmahiApi::api_key = Setting.get("api-key")
			self.install_status = 80
			uninstaller = AmahiApi::AppUninstaller.find(identifier)
			if uninstaller
				# execute the uninstall script
				self.install_status = 60
				if uninstaller.uninstall_script
					if self.webapp && self.webapp.path
						Dir.chdir(self.webapp.path) do
							SystemUtils.run_script(uninstaller.uninstall_script, name, hda_environment)
						end
					else
						Dir.chdir(app_path) do
							SystemUtils.run_script(uninstaller.uninstall_script, name, hda_environment)
						end
					end
				end
				self.install_status = 20
				self.uninstall_pkgs uninstaller if uninstaller.pkg
			# else
				# FIXME - retry? what if an app is not
				# live at this time??
			end
			# FIXME - set to nil to destroy??
			self.install_status = 0
			self.destroy
			self.install_status = nil
			FileUtils.rm_rf app_path
		rescue Exception => e
			self.install_status = 100
			FileUtils.rm_rf app_path
			raise e
		end
	end

	def theme?
		theme_id != nil
	end

	def full_url
		"http://#{app_url}.#{Setting.get 'domain'}"
	end

protected

	# extra environment for install scripts
	# *please* update the docs of variables supported at
	# http://wiki.amahi.org/index.php/Script_variables
	def hda_environment(user = nil, password = nil)
		env = {}
		net = Setting.get('net')
		addr = Setting.get('self-address')
		dom = Setting.get('domain')
		env["HDA_IP"] = [net, addr].join '.'
		env["HDA_DOMAIN"] = dom
		env["HDA_APP_DIR"] = Dir.pwd
		env["HDA_APP_NAME"] = self.name
		user && env["HDA_APP_USERNAME"] = user
		password && env["HDA_APP_PASSWORD"] = password
		env
	end

	def before_destroy_hook
		# uninstall
	end

	def mkdir path
		begin
			FileUtils.mkdir_p path
		rescue
			# maybe it exists?
		end
	end

	def online_app_to_local_app(online_app)
		name = online_app.name
		screenshot_url = online_app.screenshot_url
		identifier = online_app.id
		description = online_app.description
		version = online_app.version
		app_url = online_app.app_url
		logo_url = online_app.logo_url
		installed = false
	end

	def install_pkgs(installer)
		pkgs = installer.pkg
		return if pkgs.nil? or pkgs.blank?
		pkgs.strip!
		Platform.install(pkgs)
	end

	def install_app_deps(installer)
		deps = installer.app_dependencies
		return [] if deps.nil? or deps.blank?
		deps.strip!
		deps.split(/[, ]+/).map do |identifier|
			a = App.find_by_identifier identifier
			unless a
				a = App.new identifier
				a.install_bg
			end
			# add the dependency if it does not exist
			self.dependencies << a
		end
	end

	def install_pkg_deps(installer)
		deps = installer.pkg_dependencies
		return if deps.nil? or deps.blank?
		deps.strip!
		unless deps.blank?
			Platform.install(deps)
		end
	end

	def uninstall_pkgs(uninstaller)
		deps = uninstaller.pkg
		return if deps.nil? or deps.blank?
		deps.strip!
		unless deps.blank?
			Platform.uninstall(deps)
		end
	end

	def install_theme(installer)
		require 'sha1'

		return if (installer.source_url.nil? or installer.source_url.blank?)

    		fname = download_and_check_sha1(installer.source_url, installer.source_sha1)

		dir = nil
		Dir.chdir(File.join(RAILS_ROOT, "public/themes")) do
			mkdir '.unpack'
			Dir.chdir(".unpack") do
				unpack(installer.source_url, fname)
				# if only one file, move it to html!
				files = Dir.glob('*')
				if files.size == 1
					dir = files.first
					begin; Dir.rmdir("../#{dir}"); rescue; end
					File.rename(files.first, "../#{dir}")
				else
					# FIXME what to do if more file are unpacked
					raise "WARNING: this application unpacks into more than one file. This is a warning sign that it may not install properly!"
				end
			end
			FileUtils.rm_rf ".unpack"
		end
		Theme.dir2theme(dir)
	end

	def install_webapp(installer)
		require 'sha1'

		name = webapp_name(installer.url_name)
		path = WEBAPP_PATH % name

		# clean it first
		FileUtils.rm_rf path

		mkdir File.join(path, 'html')
		mkdir File.join(path, 'logs')

		return [name, path] if (installer.source_url.nil? or installer.source_url.blank?)

    		fname = download_and_check_sha1(installer.source_url, installer.source_sha1)

		one_dir = true
		Dir.chdir(path) do
			mkdir 'unpack'
			mkdir 'logs'
			Dir.chdir("unpack") do
				unpack(installer.source_url, fname)
				# if only one file, move it to html!
				files = Dir.glob('*')
				if files.size == 1
					Dir.rmdir("../html")
					File.rename(files.first, "../html")
				else
					# FIXME what to do if more file are unpacked
					puts "WARNING: this application unpacks into more than one file/dir. PLEASE contact the authors and well them to unpack into one dir only as most other apps so!"
					puts "NOTE: check the unpack/ directory for all the files"
					one_dir = false
				end
			end
			FileUtils.rm_rf "unpack" if one_dir
		end
		[name, path]
	end

	def unpack(url, fname)
		if (url =~ /\.zip$/)
			f = IO.popen("unzip #{fname}", "w+")
			f.close
		elsif (url =~ /\.(tar.gz|tgz)$/)
			system("tar -xzf #{fname}")
		elsif (url =~ /\.(tar.bz2)$/)
			system("tar -xjf #{fname}")
		else
			raise "File #{url} is not supported for unpacking please report it to the Amahi community!"
		end
	end

	def webapp_name(name)
		i = 0
		add = ""
		begin
			wa = Webapp.find_by_name(name + add)
			return (name+add) if wa.nil?
			raise "cannot find a suitable webapp name. giving up at #{name+add}." if i > 29
			i += 1
			add = i.to_s
		end while i < 100
	end

	def download_direct(url)
		require 'net/http'

		redirect_limit = 5
		ret = ""

		while redirect_limit > 0
			u = URI.parse(url)
			req = Net::HTTP::Get.new(u.path)
			response = Net::HTTP.start(u.host, u.port) { |http| http.request(req) }
			return response.body unless response.kind_of?(Net::HTTPRedirection)
			# it's a redirection
			location = response['location']
			old_url = url
			url = location.nil? ? (response.body.match(/<a href=\"([^>]+)\">/i)[1]) : location
			puts "NOTICE: redirected '#{old_url}' --> '#{url}' ..."
			redirect_limit -= 1
		end
		# reached end of redirect limit]!
		raise TooManyRedirects, "#{url}"
	end

	# try to download from the original download site. if that fails,
	# then try to download from amahi's mirror
	def download(url, filename, sha1)
		new_sha1 = "badbadsha1"
		begin
			file = download_direct(url)
			f = open(filename, "w")
			f.write file
			f.close
			puts "NOTE: file #{filename} written in cache"

			new_sha1 = Digest::SHA1.hexdigest(file)
		rescue => e
			puts "WARNING: primary downloaded of #{url} barfed with exception \"#{e.inspect}\""
		end

		if new_sha1 != sha1
			puts "WARNING: primary downloaded file #{filename} did not pass signature check - got #{new_sha1}, expected #{sha1}"
			new_url = File.join(AMAHI_DOWNLOAD_CACHE_SITE, sha1)
			file = download_direct(new_url)
			f = open(filename, "w")
			f.write file
			f.close
			puts "NOTE: new file #{filename} from Amahi's cache written in the cache"

			new_sha1 = Digest::SHA1.hexdigest(file)
			raise SHA1VerificationFailed, "#{new_url} (from original #{url}), '#{new_sha1}' vs. '#{sha1}' " if new_sha1 != sha1
		end
	end

	# download a file, but check in the cache first, also
	# check the sha1
	def download_and_check_sha1(url, sha1)

		mkdir HDA_DOWNLOAD_CACHE
		raise SHA1VerificationFailed, "#{url}, sha1sum provided is empty!" unless sha1
		cached_filename = File.join(HDA_DOWNLOAD_CACHE, sha1)
		if File.exists?(cached_filename)
			file = nil
			open cached_filename do |f|
				file = f.read
			end
			new_sha1 = Digest::SHA1.hexdigest(file)
			if new_sha1 == sha1
				puts "NOTE: file #{cached_filename} picked up from cache."
				FileUtils.touch cached_filename
				# return the file name, not the data
				return cached_filename
			else
				puts "WARNING: file #{cached_filename} in cache was found to be corrupted! Discarding it."
			end
		end

		# download if the above fails, i.e. no cached file OR the sha1sum failed!
		download(url, cached_filename, sha1)

		# return the file name, not the data
		cached_filename
	end
end
