# Amahi Home Server
# Copyright (C) 2007-2010 Amahi
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License v3
# (29 June 2007), as published in the COPYING file.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# file COPYING for more details.
# 
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the Amahi
# team at http://www.amahi.org/ under "Contact Us."

require 'tempfile'
require 'platform'

class Server < ActiveRecord::Base

	# CAUTION - this class *assumes* new servers are created
	# started and enabled at boot time as services!

	PID_PATH = "/var/run"
	INIT_PATH = "/etc/init.d"
	MONIT_PATH = "/etc/monit.d"
	MONIT_CONF = "/etc/monit.conf"

	validates_uniqueness_of :name
	validates_presence_of :name

	before_save  :before_save_hook
	after_create  :create_hook
	after_destroy  :destroy_hook

	def self.create_default_servers
		Server.create(:name => 'httpd', :pidfile => Platform.file_name(:apache2pid), :comment => I18n.t('apache_web_server'))
		Server.create(:name => 'named', :pidfile => 'named/named.pid', :comment => I18n.t('dns_server'))
		Server.create(:name => 'dhcpd', :comment => I18n.t('dhcp_server'))
		Server.create(:name => 'mysqld', :pidfile => 'mysqld/mysqld.pid', :comment => I18n.t('mysql_database_server'))
		Server.create(:name => 'smb', :pidfile => 'smbd.pid', :comment => I18n.t('file_server_samba'))
		Server.create(:name => 'hdactl', :comment => I18n.t('amahi_dyndns_updater'))
		Server.create(:name => 'openvpn', :pidfile => 'openvpn/amahi.pid', :comment => I18n.t('openvpn_server'))
		Server.create(:name => 'amahi-ushare', :comment => I18n.t('ushare_upnp_media_server')) if File.exists?("/var/run/amahi-ushare.pid")
		Server.create(:name => 'amahi-transmission', :comment => I18n.t('transmission_bittorrent_server')) if File.exists?("/var/run/amahi-transmission.pid")
		Server.create(:name => 'greyhole', :comment => 'Greyhole')
		r = "# WARNING - This file was automatically generated on #{Time.now}\n" \
			"\nset daemon 30\n" \
			"include #{MONIT_PATH}/logging\n" \
			"include #{MONIT_PATH}/*.conf\n"
		fname = "/tmp/server-conf-%d.%d" % [$$, rand(9999)]
		f = File.new fname, "w"
		f.write r
		c = Command.new("cp -f \"#{f.path}\" #{MONIT_CONF}")
		c.submit("rm -f \"#{f.path}\"")
		c.submit("chmod 644 /var/log/monit")
		c.submit("service monit condrestart")
		c.execute
		f.close
	end

	def pids
		# first check the pid file
		pf = pid_file
		r = []
		begin
			if File.exists?(pf) && File.readable?(pf)
				open(pf) do |p|
					list = p.readlines.map{ |line| line.gsub(/\n/, '').split(' ') }.flatten
					r = list.map{|pid| File.exists?("/proc/#{pid}") ? pid : nil }.compact
				end
				return r
			end
		rescue
			# something went wrong
			# go on, nothing to see here - some pid files may not be readable
		end

		# use pgrep instead
		begin
			p = IO.popen("pgrep #{name}")
			r = p.readlines.map {|pid| pid.gsub(/\n/, '')}
			p.close
			return r
		rescue
			[]
		end
	end

	def do_start
		c = Command.new start_cmd
		c.execute
	end

	def do_stop
		c = Command.new stop_cmd
		c.execute
	end

	def do_restart
		c = Command.new stop_cmd
		c.submit start_cmd
		c.execute
	end

protected

	def pid_file
		tmp = self.pidfile || (self.name + ".pid")
		(tmp =~ /^\//) ? tmp : File.join(PID_PATH, tmp)
	end

	def start_cmd
		tmp = self.start || (self.name + " start")
		(tmp =~ /^\//) ? tmp : File.join(INIT_PATH, tmp)
	end

	def stop_cmd
		tmp = self.stop || (self.name + " stop")
		(tmp =~ /^\//) ? tmp : File.join(INIT_PATH, tmp)
	end

	def destroy_hook
		c = Command.new("rm -f #{File.join(MONIT_PATH, name)}.conf")
		c.submit("service monit condrestart")
		c.submit stop_cmd
		c.execute
	end

	def cmd_file
		"# WARNING - This file was automatically generated on #{Time.now}\n" \
		"check process #{self.name} with pidfile #{self.pid_file}\n"	\
        	"\tstart program = \"#{self.start_cmd}\"\n"			\
        	"\tstop  program = \"#{self.stop_cmd}\"\n"
	end

	def monit_file_add
		fname = "/tmp/server-#{self.name}-%d.%d" % [$$, rand(9999)]
		f = File.new fname, "w"
		f.write cmd_file
		c = Command.new "cp -f \"#{f.path}\" #{File.join(MONIT_PATH, self.name)}.conf"
		c.submit "rm -f \"#{f.path}\""
		c.submit "service monit condrestart"
		c.execute
		f.close
	end

	def monit_file_remove
		c = Command.new("rm -f #{File.join(MONIT_PATH, name)}.conf")
		c.submit "service monit condrestart"
		c.execute
	end

	def service_on
		c = Command.new "chkconfig #{name} on"
		c.execute
	end

	def service_off
		c = Command.new "chkconfig #{name} off"
		c.execute
	end

	def before_save_hook
		if monitored_changed?
			monitored ? monit_file_add : monit_file_remove
			# DEBUG RAILS_DEFAULT_LOGGER.info "* * * MONITORED CHANGED to #{monitored}"
		end
		if start_at_boot_changed?
			start_at_boot ? service_on : service_off
			# DEBUG RAILS_DEFAULT_LOGGER.info "* * * START_AT_BOOT CHANGED to #{start_at_boot}"
		end
	end

	def create_hook
		monit_file_add
	end

end
