# Amahi Home Server
# Copyright (C) 2007-2010 Amahi
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License v3
# (29 June 2007), as published in the COPYING file.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# file COPYING for more details.
# 
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the Amahi
# team at http://www.amahi.org/ under "Contact Us."

require 'command'
require 'platform'

class Share < ActiveRecord::Base
	include Greyhole

	DEFAULT_SHARES_ROOT = '/var/hda/files'

	SIGNATURE = "Amahi configuration"
	DEFAULT_SHARES = [ "books", "pictures", "movies", "music", "docs" ].each {|s| I18n.t s }
	PDC_SETTINGS = "/var/hda/domain-settings"

	named_scope :in_disk_pool, :conditions => [ "disk_pool_copies > 0" ]

	has_many :cap_accesses, :dependent => :destroy
	has_many :users_with_share_access, :through => :cap_accesses, :source => :user

	has_many :cap_writers, :dependent => :destroy
	has_many :users_with_write_access, :through => :cap_writers, :source => :user

	before_save :before_save_hook
	before_destroy :before_destroy_hook
	after_save :after_save_hook
	after_destroy :after_destroy_hook

	validates_uniqueness_of :comment
	validates_presence_of :comment
	validates_presence_of :path

	# return the full path of a share (even if it does not exist!).
	# (this is for encapsulation purposes mostly)
	def self.full_path(path = '')
		File.join(DEFAULT_SHARES_ROOT, path)
	end

	# save the samba config file
	def self.push_shares
		domain = Setting.get "domain"
		debug = Setting.shares.find_by_name('debug').value == '1'

		smbconf = "/tmp/shares-%d.%d" % [$$, rand(99999)]
		s = File.new smbconf, "w"
		s.write(self.samba_conf(domain))
		s.flush
		s.close

		lmhosts = "/tmp/lmhosts-%d.%d" % [$$, rand(99999)]
		l = File.new lmhosts, "w"
		l.write(self.samba_lmhosts(domain))
		l.flush
		l.close

		# copy files to the samba area
		time = Time.now
		c = Command.new
		c.submit("cp /etc/samba/smb.conf \"/tmp/smb.conf.#{time}\"") if debug
		c.submit("cp \"#{smbconf}\" /etc/samba/smb.conf")
		c.submit("rm -f \"#{smbconf}\"")

		c.submit("cp /etc/samba/lmhosts \"/tmp/lmhosts.#{time}\"") if debug
		c.submit("cp \"#{lmhosts}\" /etc/samba/lmhosts")
		c.submit("rm -f \"#{lmhosts}\"")

		# reload the server
		Platform.reload(:smb)
		Platform.reload(:nmb)
		c.execute
	end

	def self.create_default_shares
		DEFAULT_SHARES.each do |s|
		    sh = Share.new
		    sh.path = Share.full_path(s)
		    sh.comment = s.capitalize
		    sh.rdonly = false
		    sh.visible = true
		    sh.tags = s
		    sh.extras = ""
		    sh.disk_pool_copies = 0
		    sh.save
		    sh.reload
		end
	end

	# configuration for one share
	def share_conf
		ret = "[%s]\n" 			\
			"\tcomment = %s\n" 	\
			"\tpath = %s\n" 	\
			"\twriteable = %s\n" 	\
			"\tbrowseable = %s\n%s%s%s%s\n"
		wr = rdonly ? "no" : "yes"
		br = visible ? "yes" : "no"
		allowed  = ''
		writes  = ''
		masks = "\tcreate mask = 0775\n"
		masks += "\tforce create mode = 0664\n"
		masks += "\tdirectory mask = 0775\n"
		masks += "\tforce directory mode = 0775\n"
		unless everyone
			allowed = "\tvalid users = "
			writes = "\twrite list = "
			u = users_with_share_access.map{ |acc| acc.login } rescue nil
			w = users_with_write_access.select{ |wrt| u.include?(wrt.login) }.map{ |user| user.login } rescue nil
			u = ['nobody'] if !u or u.empty?
			u |= ['nobody'] if guest_access
			allowed += u.join(', ') + "\n"
			w = ['nobody'] if !w or w.empty?
			w |= ['nobody'] if guest_writeable
			writes += w.join(', ') + "\n"
		end
		if (guest_access || guest_writeable) && !everyone
			writes += "\tguest ok = yes\n"
		end
		e = ""
		e = "\t" + (extras.gsub /\n/, "\n\t") unless extras.nil?
		if disk_pool_copies > 0
			tmp = e.gsub /\tdfree command.*\n/, ''
			e = tmp.gsub /\tvfs objects.*greyhole.*\n/, ''
			e += "\n\t" + 'dfree command = /usr/bin/greyhole-dfree' + "\n"
			e += "\t" + 'vfs objects = greyhole' + "\n"
		end
		ret % [comment, comment, path, wr, br, allowed, writes, masks, e]
	end

	def tag_list
		parse_tags tags
	end

	def self.basenames
		all.map { |s| [s.path, s.comment] }
	end

	def make_guest_writeable
		c = Command.new
		c.submit("chmod o+w \"#{self.path}\"")
		c.execute
	end

	def make_guest_non_writeable
		c = Command.new
		c.submit("chmod o-w \"#{self.path}\"")
		c.execute
	end

private

	def parse_tags(string)
	  string.split(/\s*,\s*|\s+/).reject {|s| s.empty? }
	end

	def before_save_hook
		return unless self.path_changed?
		return if self.path.nil? or self.path.blank?
		c = Command.new
		c.submit("rmdir \"#{self.path_was}\"") unless self.path_was.blank?
		c.submit("mkdir -p \"#{self.path}\"")
		c.submit("chown 500:100 \"#{self.path}\"")
		c.submit("chmod g+w \"#{self.path}\"")
		c.execute
	end

	def after_save_hook
		if guest_writeable_changed?
			guest_writeable ? make_guest_writeable : make_guest_non_writeable
		end
		Share.push_shares
		Greyhole.save_conf_file(DiskPoolPartition.all, Share.in_disk_pool)
	end

	def before_destroy_hook
		c = Command.new("rmdir --ignore-fail-on-non-empty \"#{self.path}\"")
		c.execute
	end

	def after_destroy_hook
		Share.push_shares
		# sync the gh configuration
		Greyhole.save_conf_file(DiskPoolPartition.all, Share.in_disk_pool)
	end

	def self.samba_conf(domain)
		ret = self.header(domain)
		Share.find(:all).each do |s|
			ret += s.share_conf
		end
		ret
	end

	def self.header_workgroup(domain)
		short_domain = Setting.shares.get "workgroup"
		debug = Setting.shares.find_by_name('debug').value == '1'
		win98 = Setting.shares.find_by_name('win98').value == '1'
		ret = ["# This file is automatically generated for WORKGROUP setup.",
		"# Any manual changes MAY BE OVERWRITTEN\n# #{SIGNATURE}, generated on #{Time.now}",
		"[global]",
		"\tworkgroup = %s",
		"\tserver string = %s",
		"\tnetbios name = hda",
		"\tprinting = cups",
		"\tprintcap name = cups",
		"\tload printers = yes",
		"\tcups options = raw",
		"\tlog file = /var/log/samba/%%m.log",
		"\tlog level = #{debug ? 5 : 0}",
		"\tmax log size = 150",
		"\tsocket options = TCP_NODELAY",
		"\tpreferred master = yes",
		"\tos level = 60",
		"\ttime server = yes",
		"\tunix extensions = no",
		"\twide links = yes",
		"\tveto files = /*.nws/riched20.dll/*.{*}/",
		"\tsecurity = user",
		"\tusername map script = /usr/share/hda-platform/hda-usermap",
		"\tlarge readwrite = yes",
		"\tencrypt passwords = yes",
		"\tdos charset = CP850",
		"\tunix charset = UTF8",
		"\tdisplay charset =  LOCALE",
		"\tguest account = nobody",
		"\tmap to guest = Bad User",
		"\twins support = yes",
		"\tprinter admin = root, @ntadmin, administrator",
		win98 ? "client lanman auth = yes" : "",
		"",
		"[homes]",
		"\tcomment = Home Directories",
		"\tvalid users = %%S",
		"\tbrowseable = no",
		"\twritable = yes",
		"\tcreate mask = 0644",
		"\tdirectory mask = 0755"].join "\n"
		ret % [short_domain, domain]
	end

	def self.header_pdc(domain)
		short_domain = Setting.shares.get "workgroup"
		debug = Setting.shares.find_by_name('debug').value == '1'
		admins = User.admins rescue ["no_such_user"]
		ret = ["# This file is automatically generated for PDC setup.",
		"# Any manual changes MAY BE OVERWRITTEN\n# #{SIGNATURE}, generated on #{Time.now}",
		"[global]",
		"\tworkgroup = %s",
		"\tserver string = %s",
		"\tnetbios name = hda",
		"\tprinting = cups",
		"\tprintcap name = cups",
		"\tload printers = yes",
		"\tcups options = raw",
		"\tlog file = /var/log/samba/%%m.log",
		"\tlog level = #{debug ? 5 : 0}",
		"\tmax log size = 150",
		"\tsocket options = TCP_NODELAY",
		"\tpreferred master = yes",
		"\tos level = 65",
		"\tdomain master = yes",
		"\tlocal master = yes",
		"\tadmin users = #{admins.map{|u| u.login}.join ', '}",
		"\tdomain logons = yes",
		"\tlogon path = \\\\%%L\\profiles\\%%U",
		"\tlogon drive = q:",
		"\tlogon home = \\\\%%N\\%%U",
		"\ttime server = yes",
		"\tunix extensions = no",
		"\twide links = yes",
		"\tveto files = /*.nws/riched20.dll/*.{*}/",
		"\tsecurity = user",
		"\tusername map script = /usr/share/hda-platform/hda-usermap ",
		"\tlarge readwrite = yes",
		"\tencrypt passwords = yes",
		"\tdos charset = CP850",
		"\tunix charset = UTF8",
		"\tdisplay charset =  LOCALE",
		"\tguest account = nobody",
		"\tmap to guest = Bad User",
		"\twins support = yes",
		"\tprinter admin = root, @ntadmin, administrator",
		"\tlogon script = %%U.bat",
		"\t# FIXME - is 99 (nobody) the right group?",
		"\tadd machine script = /usr/sbin/useradd -d /dev/null -g 99 -s /bin/false -M %%u",
		"",
		"[netlogon]",
		"\tpath = #{PDC_SETTINGS}/netlogon",
		"\tguest ok = yes",
		"\twritable = no",
		"\tshare modes = no",
		"",
		"[profiles]",
		"\tpath = #{PDC_SETTINGS}/profiles",
		"\twritable = yes",
		"\tbrowseable = no",
		"\tread only = no",
		"\tcreate mode = 0777",
		"\tdirectory mode = 0777",
		"",
		"[homes]",
		"\tcomment = Home Directories",
		"\tread only = no",
		"\twriteable = yes",
		"\tbrowseable = yes",
		"\tcreate mask = 0640",
		"\tdirectory mask = 0750",
		"\n"].join "\n"
		ret % [short_domain, domain]
	end

	def self.header_common
		["",
		"[print$]",
		"\tpath = /var/lib/samba/drivers",
		"\tread only = yes",
		"\tforce group = root",
		"\twrite list = @ntadmin root",
		"\tforce group = root",
		"\tcreate mask = 0664",
		"\tdirectory mask = 0775",
		"\tguest ok = yes",
		"",
		"[printers]",
		"\tpath = /var/spool/samba",
		"\twriteable = yes",
		"\tbrowseable = yes",
		"\tprintable = yes",
		"\tpublic = yes\n\n"].join("\n")
	end

	def self.header(domain)
		pdc = Setting.shares.find_by_name('pdc').value == '1'
		h = pdc ? header_pdc(domain) : header_workgroup(domain)
		h + "\n" + self.header_common
	end

	def self.create_logon_script(username)
		# do nothing for PDC
		pdc = Setting.shares.find_by_name('pdc').value == '1'
		return unless pdc
		return if File.exists?("#{PDC_SETTINGS}/netlogon/#{username}.bat")
		open("#{PDC_SETTINGS}/netlogon/#{username}.bat", "w", 0644) do |f|
			f.puts "REM Initial content generated by Amahi on #{Time.now}"
			f.puts "REM can be safely customized by Admin"
			f.puts "logon.bat"
		end
	end

	def self.samba_lmhosts(domain)
		ip = Setting.get('net') + '.' + Setting.get('self-address')
		ret = ["# This file is automatically generated. Any manual changes MAY BE OVERWRITTEN\n# #{SIGNATURE}, generated on #{Time.now}",
		"127.0.0.1 localhost",
		"%s hda",
		"%s files",
		"%s hda.%s",
		"%s files.%s"].join "\n"
		ret % [ip, ip, ip, domain, ip, domain]
	end

	def self.default_samba_domain(domain)
		d = domain.gsub /\.(com|net|org|local|co.uk|mobi|pro|info|asia|biz|..)$/, ''
		d = d.gsub /\./, '_'
		# fallback, in case too much gets chopped
		d = domain if d.size == 0
		d = d[-15..-1] if d.size > 15 
		d
	end
end
